/*
******************************************************************************************
*                                                                                        *
*    Code for                                                                            *
*    Distributed Coverage and Exploration in Unknown Non-Convex Environments             *
*    Copyright (C) 2010  Subhrajit Bhattacharya                                          *
*                                                                                        *
*    This program is free software: you can redistribute it and/or modify                *
*    it under the terms of the GNU General Public License as published by                *
*    the Free Software Foundation, either version 3 of the License, or                   *
*    (at your option) any later version.                                                 *
*                                                                                        *
*    This program is distributed in the hope that it will be useful,                     *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of                      *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                       *
*    GNU General Public License for more details <http://www.gnu.org/licenses/>.         *
*                                                                                        *
*                                                                                        *
*    Contact: subhrajit@gmail.com, http://fling.seas.upenn.edu/~subhrabh/                *
*                                                                                        *
*                                                                                        *
******************************************************************************************
*/


#include "../sbpl/headers.h"
#include "voronoi.h"




void DrawPixel(int x, int y, color3_t c, int size = 1);
void MainIteration(void);

VoronoiEnv_t theEnv;
char filename[1024];

int main(int argc, char *argv[])
{	
	theEnv.ReadEnvFile(argv[1]);
	strcpy(filename, argv[1]);
	
	// ===================================================
	
	int win;
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB); //glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowSize(theEnv.EnvWidth*theEnv.PlotScale, theEnv.EnvHeight*theEnv.PlotScale);
	win = glutCreateWindow("Voronoi");

	glClearColor(0.0,0.0,0.0,0.0);
	gluOrtho2D(0.0,theEnv.EnvWidth*theEnv.PlotScale,0.0,theEnv.EnvHeight*theEnv.PlotScale);
	glutDisplayFunc(MainIteration);
	glutIdleFunc(MainIteration);

	glutMainLoop();	
	
	return 0;
}



// =================================================

void MainIteration(void)
{
	static int CBcount = -1;
	CBcount++;
	printf("\nIteration %d\n", CBcount);
	
	// =================================================================
	// Do the computations - loop this
	
	clock_t StartTime, EndTime;
	
	printf(">>> Simulating sensor data...\n", CBcount);
	theEnv.UpdateMaps();
	
	printf(">>> Computing Tessellations and Centroids...\n");
	StartTime = clock();
		theEnv.DoExpansions();
		theEnv.CreateTessellations();
		if (theEnv.TARGET_NEIGHBOUR_GRAD)
			theEnv.ComputeTargets_TARGET_NEIGHBOUR_GRAD();
		else
			theEnv.ComputeTargets();
	EndTime = clock();
	printf("--> Time requred: %f s\n", (double)(EndTime-StartTime) / (double)CLOCKS_PER_SEC);
	
	// =================================================================
	// Do the plottings
	
	
	if (CBcount % theEnv.DRAW_INTERVAL == 0)
	{
		printf(">>> Plotting result...\n");
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

		int a, b, x, y, p, q;
		color3_t color;
		float colInt;
		
		// Draw tesselations and their boundaries
		for (x = 0; x < theEnv.EnvWidth; x++)
			for (y = 0; y < theEnv.EnvHeight; y++)
			{
				if (theEnv.occupancy_map_g[x][y] >= CELL_OCCUPIED)
				{
					color.set(0.0f,0.0f,0.0f);
					DrawPixel(x, y, color);
				}
				else if (theEnv.VoronoiTes[x][y])
				{
					if (theEnv.MAP_NO_UNCERTAINTY)
						colInt = 0.3 + 0.7 * (float)theEnv.VoronoiTes[x][y] / (float)theEnv.ROBOT_COUNT;
					else
						colInt = theEnv.entropy_smooth_map_g[x][y] + 0.3; //(float)theEnv.VoronoiTes[x][y] / (float)theEnv.ROBOT_COUNT;
					color.set(0.3f, colInt, 0.3f);
					//color.set(0.0f, 3.0*((float)theEnv.robots[0].valid_target_map[x][y]/1000), 0.0f);
					DrawPixel(x, y, color);
				}
				else
				{
					color.set(0.0f,0.0f,0.0f);
					DrawPixel(x, y, color);
				}
				
				// Tesselation boundaries
				color.set(1.0f,0.0f,0.0f);
				glColor3f(color.r,color.g,color.b);
				for (p=-1; p<=1; p++)
					for (q=-1; q<=1; q++)
					{
						if ((p==0 && q==0) || (p!=0 && q!=0) || x+p<0 || x+p>=theEnv.EnvWidth || y+q<0 || y+q>=theEnv.EnvHeight)
							continue;
						if ( theEnv.VoronoiTes[x][y] && theEnv.VoronoiTes[x+p][y+q]
								&& theEnv.VoronoiTes[x][y] != theEnv.VoronoiTes[x+p][y+q] )
						{
							glLineWidth(3.0);
							glBegin (GL_LINES);
							glVertex2f((x+(p+q)/2.0)*theEnv.PlotScale, (y+(q-p)/2.0)*theEnv.PlotScale);
							glVertex2f((x+(p-q)/2.0)*theEnv.PlotScale, (y+(q+p)/2.0)*theEnv.PlotScale);
							glEnd();
						}
					}
			}
		
		/*
		// Draw obstacles
		for (x = 0; x < theEnv.EnvWidth; x++)
			for (y = 0; y < theEnv.EnvHeight; y++)
				if (theEnv.occupancy_map_g[x][y] >= CELL_OCCUPIED)
				{
					color.set(0.0f,0.0f,0.0f);
					DrawPixel(x, y, color);
				}
		*/
				
		// Draw Robots and their traj histories
		for (a=0; a<theEnv.ROBOT_COUNT; a++)
		{
			color.set(0.7f,0.0f,0.7f);
			glColor3f(color.r,color.g,color.b);
			if (!theEnv.DONT_DRAW_HISTORY)
			{
				glLineWidth(2.0);
				glBegin (GL_LINE_STRIP);
				for (b=0; b<theEnv.robots[a].TrajHistory.size(); b++)
					glVertex2f((float)theEnv.robots[a].TrajHistory[b].x*theEnv.PlotScale, (float)theEnv.robots[a].TrajHistory[b].y*theEnv.PlotScale);
				glEnd();
			}
		
			color.set(1.0f,1.0f,0.0f);
			DrawPixel(theEnv.robots[a].coord.x, theEnv.robots[a].coord.y, color, 2);
				glLineWidth(2.0);
				glBegin (GL_LINE_LOOP);
				for (float th=0.0; th<=2.0*PI_CONST; th+=PI_CONST/20.0)
					glVertex2f( (float)(theEnv.robots[a].coord.x+10.0*cos(th))*theEnv.PlotScale , (float)(theEnv.robots[a].coord.y+10.0*sin(th))*theEnv.PlotScale );
				glEnd();
			
		}

		if (!theEnv.DONT_DRAW_TRAJECTORY)
		{
			// Draw Target points
			color.set(0.0f,0.5f,0.0f);
			for (a=0; a<theEnv.ROBOT_COUNT; a++)
				DrawPixel(theEnv.robots[a].TargetCoord.x, theEnv.robots[a].TargetCoord.y, color, 2);

			// Draw trajectories
			color.set(0.0f,0.0f,1.0f);
			for (a=0; a<theEnv.ROBOT_COUNT; a++)
				for (b=0; b<theEnv.robots[a].InverseTrajectory.size(); b++)
					DrawPixel(theEnv.robots[a].InverseTrajectory[b].x, theEnv.robots[a].InverseTrajectory[b].y, color);
		}

		glFlush();
	
		// ----------------------------------------------------
		// Capture image
		
		if (theEnv.CAPTURE_PNG_FRAMES.size()>0 && CBcount>=theEnv.CAPTURE_PNG_FRAMES[0])
		{
			theEnv.CAPTURE_PNG_FRAMES.erase(theEnv.CAPTURE_PNG_FRAMES.begin());
		
			int WIDTH, HEIGHT, BIT_R, BIT_G, BIT_B;
			GLvoid* imageData;
			CvSize size;
			IplImage* ipl_image_p;
			std::ostringstream imgfilename;
			std::ostringstream pdfconvertcommand;
		
			WIDTH = glutGet(GLUT_WINDOW_WIDTH);;
			HEIGHT = glutGet(GLUT_WINDOW_HEIGHT);
			BIT_R = sizeof(unsigned char);
			BIT_G = sizeof(unsigned char);
			BIT_B = sizeof(unsigned char);
			imageData = malloc(WIDTH*HEIGHT*(BIT_R+BIT_G+BIT_B)+1000); //Allocate memory for storing the image
			glReadPixels(0, 0, WIDTH, HEIGHT, GL_RGB, GL_UNSIGNED_BYTE, imageData); //Copy the image to the array imageData
	
			
			size.width = WIDTH;
			size.height = HEIGHT;
			ipl_image_p = cvCreateImageHeader(size, IPL_DEPTH_8U, 3);
			ipl_image_p->imageData = (char*)imageData;
			ipl_image_p->imageDataOrigin = ipl_image_p->imageData;
			ipl_image_p->origin = 1;
			
			imgfilename << "outfiles/" << filename << "_" << CBcount;
			pdfconvertcommand << "convert " << imgfilename.str().c_str() << ".png -compress zip " << imgfilename.str().c_str() << ".pdf";
			imgfilename << ".png";
			printf("--> Saving image: %s\n", imgfilename.str().c_str());
			cvSaveImage(imgfilename.str().c_str(), ipl_image_p);
			system(pdfconvertcommand.str().c_str());
			//cvReleaseImage(&ipl_image_p);
			free(imageData);
		}
		
	}
	
	// =================================================================
	// Control and move robots
	
	printf(">>> Moving robots...\n");
	theEnv.MoveRobots();
}

// -----------------------------

void DrawPixel(int x, int y, color3_t c, int size)
{
	glColor3f(c.r,c.g,c.b);
	glBegin(GL_POLYGON);
	glVertex2i((x+0.5-0.5*size)*theEnv.PlotScale, (y+0.5-0.5*size)*theEnv.PlotScale);
	glVertex2i((x+0.5+0.5*size)*theEnv.PlotScale, (y+0.5-0.5*size)*theEnv.PlotScale);
	glVertex2i((x+0.5+0.5*size)*theEnv.PlotScale, (y+0.5+0.5*size)*theEnv.PlotScale);
	glVertex2i((x+0.5-0.5*size)*theEnv.PlotScale, (y+0.5+0.5*size)*theEnv.PlotScale);
	glEnd();
}

