/*
******************************************************************************************
*                                                                                        *
*    Code for                                                                            *
*    Distributed Coverage and Exploration in Unknown Non-Convex Environments             *
*    Copyright (C) 2010  Subhrajit Bhattacharya                                          *
*                                                                                        *
*    This program is free software: you can redistribute it and/or modify                *
*    it under the terms of the GNU General Public License as published by                *
*    the Free Software Foundation, either version 3 of the License, or                   *
*    (at your option) any later version.                                                 *
*                                                                                        *
*    This program is distributed in the hope that it will be useful,                     *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of                      *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                       *
*    GNU General Public License for more details <http://www.gnu.org/licenses/>.         *
*                                                                                        *
*                                                                                        *
*    Contact: subhrajit@gmail.com, http://fling.seas.upenn.edu/~subhrabh/                *
*                                                                                        *
*                                                                                        *
******************************************************************************************
*/


#include "../sbpl/headers.h"
#include "voronoi.h"

point2D_t CreatePoint2D (int xx, int yy)
{
	point2D_t ret;
	ret.set(xx, yy);
	return ret;
}

// =================================================

void VoronoiEnv_t::MemAlloc(void)
{
	Grid2D = new int* [EnvWidth];
	VoronoiTes = new unsigned char* [EnvWidth];
	if (SHARE_MAP)
	{
		prob_sum_map_g = new float* [EnvWidth];
		measurement_count_map_g = new int* [EnvWidth];
		entropy_map_g = new float* [EnvWidth];
		entropy_smooth_map_g = new float* [EnvWidth];
		occupancy_map_g = new int* [EnvWidth];
	}
	
	for (int xx = 0; xx < EnvWidth; xx++)
	{
		Grid2D[xx] = new int [EnvHeight];
		VoronoiTes[xx] = new unsigned char [EnvHeight];
		for (int yy = 0; yy < EnvHeight; yy++)
			Grid2D[xx][yy] = CELL_NOT_OCCUPIED;
		
		if (SHARE_MAP)
		{
			prob_sum_map_g[xx] = new float [EnvHeight];
			measurement_count_map_g[xx] = new int [EnvHeight];
			entropy_map_g[xx] = new float [EnvHeight];
			entropy_smooth_map_g[xx] = new float [EnvHeight];
			occupancy_map_g[xx] = new int [EnvHeight];
			for (int yy = 0; yy < EnvHeight; yy++)
			{
				prob_sum_map_g[xx][yy] = 0.5;
				measurement_count_map_g[xx][yy] = 1;
				entropy_map_g[xx][yy] = LOG_HALF;
				entropy_smooth_map_g[xx][yy] = LOG_HALF;
				occupancy_map_g[xx][yy] = CELL_NOT_OCCUPIED + (int)(LOG_HALF*METRIC_WEIGHT_OF_ENTROPY);
			}
		}
	}
}

void VoronoiEnv_t::AddRobot(int x, int y, float sensor_Rad_sq, float sensor_Near_truth, float sensor_Far_truth)
{
	ROBOT_COUNT++;
	
	robots.resize(ROBOT_COUNT);
	robots[ROBOT_COUNT-1].coord.set(x,y);
	robots[ROBOT_COUNT-1].search_expansion = new SBPL2DGridSearch(EnvWidth, EnvHeight, 1.0);
	robots[ROBOT_COUNT-1].sensor_rad_sq = sensor_Rad_sq;
	robots[ROBOT_COUNT-1].sensor_near_truth = sensor_Near_truth;
	robots[ROBOT_COUNT-1].sensor_far_truth = sensor_Far_truth;
	robots[ROBOT_COUNT-1].recomputeTarget = true;
	
	robots[ROBOT_COUNT-1].valid_target_map = new int* [EnvWidth];
	for (int xx = 0; xx < EnvWidth; xx++)
	{
		robots[ROBOT_COUNT-1].valid_target_map[xx] = new int [EnvHeight];
	}
	
	if (SHARE_MAP)
	{
		robots[ROBOT_COUNT-1].prob_sum_map = prob_sum_map_g;
		robots[ROBOT_COUNT-1].measurement_count_map = measurement_count_map_g;
		robots[ROBOT_COUNT-1].entropy_map = entropy_map_g;
		robots[ROBOT_COUNT-1].entropy_smooth_map = entropy_smooth_map_g;
		robots[ROBOT_COUNT-1].occupancy_map = occupancy_map_g;
	}
	else
	{
		robots[ROBOT_COUNT-1].prob_sum_map = new float* [EnvWidth];
		robots[ROBOT_COUNT-1].measurement_count_map = new int* [EnvWidth];
		robots[ROBOT_COUNT-1].entropy_map = new float* [EnvWidth];
		robots[ROBOT_COUNT-1].entropy_smooth_map = new float* [EnvWidth];
		robots[ROBOT_COUNT-1].occupancy_map = new int* [EnvWidth];
		for (int xx = 0; xx < EnvWidth; xx++)
		{
			robots[ROBOT_COUNT-1].prob_sum_map[xx] = new float [EnvHeight];
			robots[ROBOT_COUNT-1].measurement_count_map[xx] = new int [EnvHeight];
			robots[ROBOT_COUNT-1].entropy_map[xx] = new float [EnvHeight];
			robots[ROBOT_COUNT-1].entropy_smooth_map[xx] = new float [EnvHeight];
			robots[ROBOT_COUNT-1].occupancy_map[xx] = new int [EnvHeight];
			for (int yy = 0; yy < EnvHeight; yy++)
			{
				robots[ROBOT_COUNT-1].prob_sum_map[xx][yy] = 0.5;
				robots[ROBOT_COUNT-1].measurement_count_map[xx][yy] = 1;
				robots[ROBOT_COUNT-1].entropy_map[xx][yy] = LOG_HALF;
				robots[ROBOT_COUNT-1].entropy_smooth_map[xx][yy] = LOG_HALF;
				robots[ROBOT_COUNT-1].occupancy_map[xx][yy] = CELL_NOT_OCCUPIED + (int)(LOG_HALF*METRIC_WEIGHT_OF_ENTROPY);
			}
		}
	}
}

// =======================================================

void VoronoiEnv_t::DoExpansions(void)
{
	for(int a=0; a<ROBOT_COUNT; a++)
		robots[a].search_expansion->search(robots[a].occupancy_map, CELL_OCCUPIED_THRESH, robots[a].occupancy_map, CELL_OCCUPIED_THRESH, robots[a].coord.x, robots[a].coord.y);
		//robots[a].search_expansion->search(robots[a].occupancy_map, CELL_OCCUPIED_THRESH, robots[a].coord.x, robots[a].coord.y, 0, 0,  SBPL_2DGRIDSEARCH_TERM_CONDITION_ALLCELLS);
}

// -----------------

void VoronoiEnv_t::CreateTessellations(void)
{
	int thisCostInt, minRob;
	float minCost, thisCost;
	int** occupancy_map_to_use;
	float** entropy_map_to_use;
	float** entropy_smooth_map_to_use;
	
	if (SHARE_MAP)
	{
		occupancy_map_to_use = occupancy_map_g;
		entropy_map_to_use = entropy_map_g;
		entropy_smooth_map_to_use = entropy_smooth_map_g;
	}
	else
	{
		occupancy_map_to_use = NULL; // Change this to something meaningful for distributed case
		entropy_map_to_use = NULL;
		entropy_smooth_map_to_use = NULL;
	}
	
	for (int xx = 0; xx < EnvWidth; xx++)
		for (int yy = 0; yy < EnvHeight; yy++)
		{
			minCost = FLOAT_INF;
			minRob = 0;
			if (!(occupancy_map_to_use[xx][yy] >= CELL_OCCUPIED))
				for(int a=0; a<ROBOT_COUNT; a++)
				{
					thisCostInt = robots[a].search_expansion->getlowerboundoncostfromstart_inmm(xx,yy);
					if (thisCostInt==INFINITECOST) // Inside an obstacle - inaccessible
					{
						minRob = 0;
						break;
					}
					thisCost = (float)thisCostInt / SBPL_2DGRIDSEARCH_COSTMULTIPLIER;
					thisCost = PowerDiagramDistance(thisCost, robots[a].sensor_rad_sq);
					if (thisCost < minCost)
					{
						minRob = a + 1;
						minCost = thisCost;
					}
				}
			VoronoiTes[xx][yy] = minRob;
		}
	
	// Create smooth entropy map and valid target map
	
	float en_sum;
	int sx, sy, en_count;
	
	for (int xx = 0; xx < EnvWidth; xx++)
		for (int yy = 0; yy < EnvHeight; yy++)
		{
			en_sum = 10.0; //0.0;
			en_count = 0;
			for (sx = __max(xx-ENTROPY_SMOOTHEN_RADIUS,0); sx < __min(xx+ENTROPY_SMOOTHEN_RADIUS,EnvWidth); sx++)
				for (sy = __max(yy-ENTROPY_SMOOTHEN_RADIUS,0); sy < __min(yy+ENTROPY_SMOOTHEN_RADIUS,EnvHeight); sy++)
					if (VoronoiTes[sx][sy] == VoronoiTes[xx][yy])
					{
						//en_sum += entropy_map_to_use[sx][sy];
						//en_count++;
						en_sum = __min(en_sum, entropy_map_to_use[sx][sy]);
					}
			//entropy_smooth_map_to_use[xx][yy] = en_sum / en_count;
			entropy_smooth_map_to_use[xx][yy] = en_sum;
		}
		
	for (int xx = 0; xx < EnvWidth; xx++)
		for (int yy = 0; yy < EnvHeight; yy++)
			for(int a=0; a<ROBOT_COUNT; a++)
			{
				if (robots[a].occupancy_map[xx][yy] == CELL_OCCUPIED || VoronoiTes[xx][yy] != a+1)
					robots[a].valid_target_map[xx][yy] = INT_MAX;
				else
					robots[a].valid_target_map[xx][yy] = (int)(entropy_smooth_map_to_use[xx][yy]*1000.0);
			}
}

// -----------------

void VoronoiEnv_t::ComputeTargets(void)
{
	vector<float> int_x_wt (ROBOT_COUNT, 0.0);
	vector<float> int_y_wt (ROBOT_COUNT, 0.0);
	vector<float> int_wt (ROBOT_COUNT, 0.0);
	vector<int> int_count (ROBOT_COUNT, 0);
	int robNo;
	float prob, phi1, phi2, wt, dx, dy;
	float** prob_sum_map_to_use;
	int** measurement_count_map_to_use;
	float** entropy_map_to_use;
	
	if (SHARE_MAP)
	{
		prob_sum_map_to_use = prob_sum_map_g;
		measurement_count_map_to_use = measurement_count_map_g;
		entropy_map_to_use = entropy_map_g;
	}
	else
	{
		prob_sum_map_to_use = NULL; // Change this to something meaningful for distributed case
		measurement_count_map_to_use = NULL; // Change this to something meaningful for distributed case
		entropy_map_to_use = NULL;
	}
	
	for(int a=0; a<ROBOT_COUNT; a++)
		if (!robots[a].recomputeTarget && 
				(VoronoiTes[robots[a].TargetCoord.x][robots[a].TargetCoord.y]!=a+1 || robots[a].TargetCoord.dist(robots[a].coord)<=2.0) )
			robots[a].recomputeTarget = true;
	
	// Computing the entropies
	for (int xx = 0; xx < EnvWidth; xx++)
		for (int yy = 0; yy < EnvHeight; yy++)
			if (VoronoiTes[xx][yy])
			{
				robNo = VoronoiTes[xx][yy] - 1;
				if (!robots[robNo].recomputeTarget)
					continue;
				
				phi1 = entropy_map_to_use[xx][yy];
				phi1 = pow(phi1, ENTROPY_WEIGHT_POWER);
				
				wt = phi1;
				
				int_x_wt[robNo] += (float)xx * wt;
				int_y_wt[robNo] += (float)yy * wt;
				int_wt[robNo] += wt;
				int_count[robNo]++;
			}
			
	for(int a=0; a<ROBOT_COUNT; a++)
		if (robots[a].recomputeTarget)
		{
			//printf("### %d - %d, %d: %f, %f, %f ; %d, %d; %d, %d \n", a, int_count[a], robots[a].recomputeTarget, int_x_wt[a], int_y_wt[a], int_wt[a], VoronoiTes[robots[a].TargetCoord.x][robots[a].TargetCoord.y], VoronoiTes[robots[a].coord.x][robots[a].coord.y], occupancy_map_g[robots[a].TargetCoord.x][robots[a].TargetCoord.y], occupancy_map_g[robots[a].coord.x][robots[a].coord.y]);
			robots[a].TargetCoord.x = (int) (int_x_wt[a] / int_wt[a]);
			robots[a].TargetCoord.y = (int) (int_y_wt[a] / int_wt[a]);
			//robots[a].TargetCoord = robots[a].coord;
			if (!MAP_NO_UNCERTAINTY)
				robots[a].recomputeTarget = false;
		}
	
	for(int a=0; a<ROBOT_COUNT; a++)
	{
			robots[a].TargetCoord = ProjectPointInsideTessellation(robots[a].TargetCoord, a);
			robots[a].TargetCoord = ProjectPointInsideHighEntropy(robots[a].TargetCoord, a, TARGET_CELL_MIN_ENTROPY);
	}

}

//---------------------------------
// Experimental - only for 8-connected graph
void VoronoiEnv_t::ComputeTargets_TARGET_NEIGHBOUR_GRAD(void)
{
	SBPL2DGridSearch* search_expansion;
	bool newAlloc = false;
	float** entropy_map_to_use;
	int** tmpOccupancyMap;
	tmpOccupancyMap = new int* [EnvWidth];
	for (int xx = 0; xx < EnvWidth; xx++)
		tmpOccupancyMap[xx] = new int [EnvHeight];
	
	if (SHARE_MAP)
		entropy_map_to_use = entropy_map_g;
	else
		entropy_map_to_use = NULL;
	
	for (int a=0; a<ROBOT_COUNT; a++)
	{
		float min_int_sum = 1.0/0.0;
		
		for (int xx = 0; xx < EnvWidth; xx++)
			for (int yy = 0; yy < EnvHeight; yy++)
				if (VoronoiTes[xx][yy]-1 != a)
					tmpOccupancyMap[xx][yy] = CELL_OCCUPIED_THRESH;
				else
					tmpOccupancyMap[xx][yy] = robots[a].occupancy_map[xx][yy];
	
		for (int nx = __max(robots[a].coord.x-1,0); nx <= __min(robots[a].coord.x+1,EnvWidth-1); nx++)
			for (int ny = __max(robots[a].coord.y-1,0); ny <= __min(robots[a].coord.y+1,EnvHeight-1); ny++)
			{
				if (VoronoiTes[nx][ny]-1 != a)
					continue;
				
				if ((nx==robots[a].coord.x && nx==robots[a].coord.y))
				{
					search_expansion = robots[a].search_expansion;
					newAlloc = false;
				}
				else
				{
					search_expansion = new SBPL2DGridSearch(EnvWidth, EnvHeight, 1.0);;
					search_expansion->search(tmpOccupancyMap, CELL_OCCUPIED_THRESH, tmpOccupancyMap, CELL_OCCUPIED_THRESH, nx, ny);
					newAlloc = true;
				}
				
				float int_sum = 0.0;
				for (int xx = 0; xx < EnvWidth; xx++)
					for (int yy = 0; yy < EnvHeight; yy++)
						if (VoronoiTes[xx][yy]-1 == a)
							int_sum += pow(((float)search_expansion->getlowerboundoncostfromstart_inmm(xx,yy))/1000.0, 2) 
										* pow(entropy_map_to_use[xx][yy], ENTROPY_WEIGHT_POWER);
				
				if (int_sum <= min_int_sum)
				{
					min_int_sum = int_sum;
					robots[a].TargetCoord = CreatePoint2D(nx,ny);
				}
				
				if (newAlloc)
					delete search_expansion;
			}
	}
	
	for (int xx = 0; xx < EnvWidth; xx++)
		delete[] tmpOccupancyMap[xx];
	delete[] tmpOccupancyMap;
}
//---------------------------------

point2D_t VoronoiEnv_t::ProjectPointInsideTessellation(point2D_t pt, int rob_index)
{
	point2D_t projectedPt = pt;
	bool centroidOK;
	int tx, ty;
	float r, th;
	
	// Project to the same tessellation as the robot
	if (VoronoiTes[projectedPt.x][projectedPt.y] != rob_index+1)
	{
		centroidOK = false;
		for (r=1.0; r<=__max(EnvWidth,EnvHeight); r++)
		{
			for (th=-PI_CONST/2.0; th<=PI_CONST/2.0; th+=1.0/r)
			{
				tx = pt.x + (int)(r*cos(th));
				ty = pt.y + (int)(r*sin(th));
				if(tx>=0 && ty >=0 && tx<EnvWidth && ty<EnvHeight) 
					if (VoronoiTes[tx][ty] == rob_index+1)
					{
						projectedPt.x = tx;
						projectedPt.y = ty;
						centroidOK = true;
						break;
					}
			}
			if (centroidOK)
				break;
		}
	}
	
	return projectedPt;
}


point2D_t VoronoiEnv_t::ProjectPointInsideHighEntropy(point2D_t pt, int rob_index, float entropy_thresh)
{
	point2D_t projectedPt = pt;
	float** entropy_smooth_map_to_use;
	
	if (SHARE_MAP)
		entropy_smooth_map_to_use = entropy_smooth_map_g;
	else
		entropy_smooth_map_to_use = NULL;
	
	// Project to nearest high entropy region
	if (entropy_smooth_map_to_use[projectedPt.x][projectedPt.y] <= entropy_thresh)
	{
		SBPL2DGridSearch* search_expansion = new SBPL2DGridSearch(EnvWidth, EnvHeight, 1.0);
		SBPL_2DGridSearchState goalState = search_expansion->search(robots[rob_index].valid_target_map, INT_MAX, robots[rob_index].valid_target_map, 0, projectedPt.x, projectedPt.y, robots[rob_index].valid_target_map, (int)(entropy_thresh*1000.0));
		projectedPt.x = goalState.x;
		projectedPt.y = goalState.y;
		delete search_expansion;
	}
	
	return projectedPt;
}

// -----------------

void VoronoiEnv_t::UpdateMaps(void)
{
	int** occupancy_map_to_use;
	float** prob_sum_map_to_use;
	int** measurement_count_map_to_use;
	float** entropy_map_to_use;
	float prob_true, rob_dist_sq, DX, DY, prob, entropy;
	bool true_occupied, mm, recompute_occupancy, robot_cell;
	//float max_entropy = 0.0;

	if (SHARE_MAP)
	{
		occupancy_map_to_use = occupancy_map_g;
		prob_sum_map_to_use = prob_sum_map_g;
		measurement_count_map_to_use = measurement_count_map_g;
		entropy_map_to_use = entropy_map_g;
	}
	else
	{
		occupancy_map_to_use = NULL; // Change this to something meaningful for distributed case
		prob_sum_map_to_use = NULL; // Change this to something meaningful for distributed case
		measurement_count_map_to_use = NULL; // Change this to something meaningful for distributed case
		entropy_map_to_use = NULL;
	}
	
	if (MAP_NO_UNCERTAINTY)
	{
		for (int xx = 0; xx < EnvWidth; xx++)
			for (int yy = 0; yy < EnvHeight; yy++)
			{
				if (Grid2D[xx][yy] >= CELL_OCCUPIED)
				{
					prob_sum_map_to_use[xx][yy] = 1.0;
					occupancy_map_to_use[xx][yy] = CELL_OCCUPIED;
				}
				else
				{
					prob_sum_map_to_use[xx][yy] = 0.0;
					occupancy_map_to_use[xx][yy] = CELL_NOT_OCCUPIED;
				}
				measurement_count_map_to_use[xx][yy] = 1;
				entropy_map_to_use[xx][yy] = ENTROPY_THRESH_FOR_UNCERTAINTY;
			}
		return;
	}
	
	// Update the probability map - simulation of sensing
	for (int xx = 0; xx < EnvWidth; xx++)
		for (int yy = 0; yy < EnvHeight; yy++)
		{
			if (Grid2D[xx][yy] >= CELL_OCCUPIED)
				true_occupied = true;
			else
				true_occupied = false;
			
			recompute_occupancy = false;
			for(int a=0; a<ROBOT_COUNT; a++)
			{
				if (IsCellVisibleToRobot(xx, yy, a))
				{
					DX = (float)(xx - robots[a].coord.x);
					DY = (float)(yy - robots[a].coord.y);
					rob_dist_sq = DX*DX + DY*DY;
					prob_true = SensorProbabilityOfTrue(rob_dist_sq, robots[a].sensor_rad_sq, 
					                                    robots[a].sensor_near_truth, robots[a].sensor_far_truth);
					mm = SensorSample(true_occupied, prob_true);
					if (mm)
						prob_sum_map_to_use[xx][yy] += prob_true;
					else
						prob_sum_map_to_use[xx][yy] += (1.0-prob_true);
					measurement_count_map_to_use[xx][yy]++;
					recompute_occupancy = true;
				}
			}
			
			if (recompute_occupancy)
			{
				prob = prob_sum_map_to_use[xx][yy] / (float)measurement_count_map_to_use[xx][yy];
				entropy = (prob<=0.0 || prob>=1.0) ? 0.0 : -( prob*log(prob) + (1.0-prob)*log(1.0-prob) );
				if (entropy < ENTROPY_THRESH_FOR_UNCERTAINTY)
					entropy = ENTROPY_THRESH_FOR_UNCERTAINTY;
				entropy_map_to_use[xx][yy] = entropy;
					
				if ( prob < P_THRESH_OBS)
					occupancy_map_to_use[xx][yy] = CELL_NOT_OCCUPIED + (int)(entropy*METRIC_WEIGHT_OF_ENTROPY);
				else
					occupancy_map_to_use[xx][yy] = CELL_OCCUPIED;
			}
		}
		
}

// -----------------

bool VoronoiEnv_t::IsCellVisibleToRobot(int cell_x, int cell_y, int rob_index)
{
	float DX, DY, rob_dist_sq;
	float steps, max_steps, step_size;
	int x_t, y_t;
	
	float ClosestCornDistSq = FLOAT_INF;
	float corn_x, corn_y, closest_corn_x, closest_corn_y;
	for (float p=-0.5; p<=0.5; p++)
		for (float q=-0.5; q<=0.5; q++)
		{
			corn_x = (float)cell_x + p;
			corn_y = (float)cell_y + q;
			DX = (float)robots[rob_index].coord.x - corn_x;
			DY = (float)robots[rob_index].coord.y - corn_y;
			rob_dist_sq = DX*DX + DY*DY;
			if (rob_dist_sq<ClosestCornDistSq)
			{
				closest_corn_x = corn_x;
				closest_corn_y = corn_y;
				ClosestCornDistSq = rob_dist_sq;
			}
		}
	
	DX = (float)robots[rob_index].coord.x - closest_corn_x;
	DY = (float)robots[rob_index].coord.y - closest_corn_y;
	rob_dist_sq = ClosestCornDistSq;
	if (rob_dist_sq < robots[rob_index].sensor_rad_sq)
	{
		max_steps = (int)__max(fabs(DX),fabs(DY));
		step_size = 1.0 / max_steps;
		for (steps = step_size; steps < 1.0; steps+=step_size)
		{
			x_t = round(closest_corn_x + DX*steps);
			y_t = round(closest_corn_y + DY*steps);
			if (Grid2D[x_t][y_t] >= CELL_OCCUPIED)
				return false;
		}
		return true;
	}
	else
		return false;
}


// =======================================================
// System model functions

float PowerDiagramDistance(float dist, float RadSq)
{
	return (dist*dist - RadSq);
}

float SensorProbabilityOfTrue(float DistSq, float SensorRadSq, float SensorNearTruth, float SensorFarTruth)
{
	return (SensorNearTruth - (SensorNearTruth-SensorFarTruth)*DistSq/SensorRadSq);
}

bool SensorSample(bool TrueOccupancy, float ProbabilityOfTrue)
{
	float sample;
	sample = (float)rand() / RAND_MAX;
	if (sample < ProbabilityOfTrue)
		return TrueOccupancy;
	else
		return (!TrueOccupancy);
}

// =======================================================
// Control functions

void VoronoiEnv_t::MoveRobots(void) // finds trajectories and moves robots
{
	int nx, ny, trajPtMaxIndex, trajMovePt, oldTrajLength;
	SBPL2DGridSearch* search_expansion_ct;
	search_expansion_ct = new SBPL2DGridSearch(EnvWidth, EnvHeight, 1.0);
	
	for(int a=0; a<ROBOT_COUNT; a++)
	{
		robots[a].TrajHistory.push_back(robots[a].coord);
		
		if (TARGET_NEIGHBOUR_GRAD)
		{
			robots[a].InverseTrajectory.clear();
			robots[a].InverseTrajectory.push_back(robots[a].coord);
			robots[a].InverseTrajectory.push_back(robots[a].TargetCoord);
			robots[a].coord = robots[a].TargetCoord;
			continue;
		}

		oldTrajLength = robots[a].InverseTrajectory.size();
		robots[a].InverseTrajectory.clear();
		robots[a].InverseTrajectory.push_back(robots[a].coord);
		
		search_expansion_ct->search(robots[a].valid_target_map, INT_MAX, robots[a].valid_target_map, 0, robots[a].TargetCoord.x, robots[a].TargetCoord.y);
	
		//search_expansion->getLowestNeighbour(robots[a].coord.x, robots[a].coord.y, &nx, &ny);
		//robots[a].coord.x = nx;
		//robots[a].coord.y = ny;
		
		int count=0;
		do
		{
			trajPtMaxIndex = robots[a].InverseTrajectory.size() - 1;
			search_expansion_ct->getLowestNeighbour(robots[a].InverseTrajectory[trajPtMaxIndex].x, robots[a].InverseTrajectory[trajPtMaxIndex].y, &nx, &ny, true);
			robots[a].InverseTrajectory.push_back(CreatePoint2D(nx,ny));
			count++;
		}
		while ( !( nx==robots[a].TargetCoord.x && ny==robots[a].TargetCoord.y ) && count<=EnvWidth+EnvHeight);

	
		trajMovePt = __min(1, robots[a].InverseTrajectory.size()-1); //__max(robots[a].InverseTrajectory.size()-2, 0);
		robots[a].coord = robots[a].InverseTrajectory[trajMovePt];
		
		if ((float)robots[a].InverseTrajectory.size()/(float)oldTrajLength > 2.0)
			robots[a].recomputeTarget = true;
	}
	
	delete search_expansion_ct;
}

// =======================================================

void VoronoiEnv_t::ReadEnvFile(const char* sEnvFile)
{
	char thisLine[1024], firstString[1024], checkString[1024];
	char longstring[4096];
	char s1[1024];
	int i1, i2, i3, i4, i5, i6, i7;
	float f1, f2, f3, f4, f5, f6, f7;
	int a, b, c, d;
	int InfRad=0;
	
	bool FileEnd=false;
	bool insideSTATICOBSTACLE = false;
	bool insideROBOT = false;
	bool insideBLOCKCOMMENT = false;

	SHARE_MAP = false;
	MAP_NO_UNCERTAINTY = false;
	TARGET_NEIGHBOUR_GRAD = false;
	DONT_DRAW_TRAJECTORY = false;
	DONT_DRAW_HISTORY = false;
	P_THRESH_OBS = 0.6;
	ENTROPY_WEIGHT_POWER = 1.0;
	ENTROPY_THRESH_FOR_UNCERTAINTY = 0.01;
	TARGET_CELL_MIN_ENTROPY = 0.0;
	ENTROPY_SMOOTHEN_RADIUS = 3;
	DRAW_INTERVAL = 2;

	FILE* fCfg = fopen(sEnvFile, "r");
	if(fCfg == NULL)
	{
		printf("ERROR: unable to open %s\n", sEnvFile);
		exit(1);
	}
	
	while(!FileEnd)
	{
		strcpy(thisLine,"");
		fgets(thisLine, 1024 , fCfg);
		if(strlen(thisLine)<=1)
			continue;
		sscanf(thisLine, "%s", firstString);
		
		// BLOCK_COMMENT_BEGIN
		strcpy(checkString, "/*");
		if(strcmp(checkString, firstString) == 0)
		{
			insideBLOCKCOMMENT = true;
			continue;
		}
		
		// BLOCK_COMMENT_END
		strcpy(checkString, "*/");
		if(strcmp(checkString, firstString) == 0 && insideBLOCKCOMMENT)
		{
			insideBLOCKCOMMENT = false;
			continue;
		}
		
		if (insideBLOCKCOMMENT)
			continue;
		
		// ---------------------------------------------------
		
		// SHARE_MAP
		strcpy(checkString, "SHARE_MAP");
		if(strcmp(checkString, firstString) == 0)
		{
			SHARE_MAP = true;
			continue;
		}
		
		// MAP_NO_UNCERTAINTY
		strcpy(checkString, "MAP_NO_UNCERTAINTY");
		if(strcmp(checkString, firstString) == 0)
		{
			MAP_NO_UNCERTAINTY = true;
			continue;
		}
		
		// TARGET_NEIGHBOUR_GRAD
		strcpy(checkString, "TARGET_NEIGHBOUR_GRAD");
		if(strcmp(checkString, firstString) == 0)
		{
			TARGET_NEIGHBOUR_GRAD = true;
			continue;
		}
		
		// DONT_DRAW_TRAJECTORY
		strcpy(checkString, "DONT_DRAW_TRAJECTORY");
		if(strcmp(checkString, firstString) == 0)
		{
			DONT_DRAW_TRAJECTORY = true;
			continue;
		}
		
		// DONT_DRAW_HISTORY
		strcpy(checkString, "DONT_DRAW_HISTORY");
		if(strcmp(checkString, firstString) == 0)
		{
			DONT_DRAW_HISTORY = true;
			continue;
		}
		
		// DRAW_INTERVAL
		strcpy(checkString, "DRAW_INTERVAL:");
		if(strcmp(checkString, firstString) == 0)
		{
			sscanf(thisLine, "%*s %d", &i1);
			DRAW_INTERVAL = i1;
			continue;
		}
		
		// CAPTURE_PNG
		strcpy(checkString, "CAPTURE_PNG:");
		if(strcmp(checkString, firstString) == 0)
		{
			sscanf(thisLine, "%*s %s", longstring);
			int commapos = 0, found;
			string frames = longstring;
			while (commapos < frames.size())
			{
				found = frames.find(",", commapos);
				if (found!=string::npos)
				{
					CAPTURE_PNG_FRAMES.push_back(atoi(frames.substr(commapos,found-commapos).c_str()));
					commapos = found+1;
				}
				else
				{
					CAPTURE_PNG_FRAMES.push_back(atoi(frames.substr(commapos).c_str()));
					break;
				}
			}
			continue;
		}
		
		// PROBABILITY_THRESH_FOR_OBSTACLE
		strcpy(checkString, "PROBABILITY_THRESH_FOR_OBSTACLE:");
		if(strcmp(checkString, firstString) == 0)
		{
			sscanf(thisLine, "%*s %f", &f1);
			P_THRESH_OBS = f1;
			continue;
		}
		
		// ENTROPY_WEIGHT_POWER
		strcpy(checkString, "ENTROPY_WEIGHT_POWER:");
		if(strcmp(checkString, firstString) == 0)
		{
			sscanf(thisLine, "%*s %f", &f1);
			ENTROPY_WEIGHT_POWER = f1;
			continue;
		}
		
		// ENTROPY_SMOOTHEN_RADIUS
		strcpy(checkString, "ENTROPY_SMOOTHEN_RADIUS:");
		if(strcmp(checkString, firstString) == 0)
		{
			sscanf(thisLine, "%*s %d", &i1);
			ENTROPY_SMOOTHEN_RADIUS = i1;
			continue;
		}
		
		// ENTROPY_THRESH_FOR_UNCERTAINTY
		strcpy(checkString, "ENTROPY_THRESH_FOR_UNCERTAINTY:");
		if(strcmp(checkString, firstString) == 0)
		{
			sscanf(thisLine, "%*s %f", &f1);
			ENTROPY_THRESH_FOR_UNCERTAINTY = f1;
			continue;
		}
		
		// TARGET_CELL_MIN_ENTROPY
		strcpy(checkString, "TARGET_CELL_MIN_ENTROPY:");
		if(strcmp(checkString, firstString) == 0)
		{
			sscanf(thisLine, "%*s %f", &f1);
			TARGET_CELL_MIN_ENTROPY = f1;
			continue;
		}
		
		// METRIC_WEIGHT_OF_ENTROPY
		strcpy(checkString, "METRIC_WEIGHT_OF_ENTROPY:");
		if(strcmp(checkString, firstString) == 0)
		{
			sscanf(thisLine, "%*s %f", &f1);
			METRIC_WEIGHT_OF_ENTROPY = f1;
			continue;
		}
		
		// +++++++++++++++++++++++++++++++++++++++++++
		
		// discretization
		strcpy(checkString, "discretization:");
		if(strcmp(checkString, firstString) == 0)
		{
			sscanf(thisLine, "%*s %d %d", &i1, &i2);
			EnvWidth = i1;
			EnvHeight = i2;
			MemAlloc();
			PlotScale = 500.0 / (float)__max(EnvWidth,EnvHeight);
			continue;
		}
		
		// scale
		strcpy(checkString, "scale:");
		if(strcmp(checkString, firstString) == 0)
		{
			sscanf(thisLine, "%*s %f", &f1);
			LENGTH_SCALE = f1;
			continue;
		}
		
		// ROBOT
		// +++++++++++++++++++++++++++++++++++++++++++
		
		strcpy(checkString, "BEGIN_ROBOT");
		if(strcmp(checkString, firstString) == 0)
		{
			// Reserves i1, i2, f1, f2, f3
			insideROBOT = true;
			continue;
		}
		
		// start
		strcpy(checkString, "start:");
		if(strcmp(checkString, firstString) == 0 && insideROBOT)
		{
			sscanf(thisLine, "%*s %f %f", &f4, &f5);
			i1 = Length2Discritization_i(f4);
			i2 = Length2Discritization_i(f5);
			continue;
		}
		
		// sensor-radius
		strcpy(checkString, "sensor-radius:");
		if(strcmp(checkString, firstString) == 0 && insideROBOT)
		{
			sscanf(thisLine, "%*s %f", &f1);
			f1 = Length2Discritization_f(f1);
			continue;
		}
		
		// sensor-truth
		strcpy(checkString, "sensor-truth:");
		if(strcmp(checkString, firstString) == 0 && insideROBOT)
		{
			sscanf(thisLine, "%*s %f %f", &f2, &f3);
			continue;
		}
		
		// END
		strcpy(checkString, "END_ROBOT");
		if(strcmp(checkString, firstString) == 0 && insideROBOT)
		{
			insideSTATICOBSTACLE = false;
			AddRobot(i1, i2, f1*f1, f2, f3);
			insideROBOT = false;
			continue;
		}
		
		// STATICOBSTACLE
		// +++++++++++++++++++++++++++++++++++++++++++
		
		strcpy(checkString, "BEGIN_STATICOBSTACLE");
		if(strcmp(checkString, firstString) == 0)
		{
			insideSTATICOBSTACLE = true;
			continue;
		}
		
		// inflate
		strcpy(checkString, "inflate:");
		if(strcmp(checkString, firstString) == 0)
		{
			sscanf(thisLine, "%*s %d", &i1);
			InfRad = i1;
			continue;
		}
		
		// rectangle
		strcpy(checkString, "rectangle:");
		if(strcmp(checkString, firstString) == 0 && insideSTATICOBSTACLE)
		{
			sscanf(thisLine, "%*s %f %f %f %f", &f1, &f2, &f3, &f4);
			i1 = Length2Discritization_i(f1);
			i2 = Length2Discritization_i(f2);
			i3 = Length2Discritization_i(f3);
			i4 = Length2Discritization_i(f4);
			for(a = __max(__min(i1,i3),InfRad); a <= __min(__max(i1,i3),EnvWidth-1-InfRad); a++)
				for(b = __max(__min(i2,i4),InfRad); b <= __min(__max(i2,i4),EnvHeight-1-InfRad); b++)
						for(c = -InfRad; c<=InfRad; c++)
							for(d = -InfRad; d<=InfRad; d++)
								Grid2D[a+c][b+d] = CELL_OCCUPIED;
		}
		
		// image
		strcpy(checkString, "image:");
		if(strcmp(checkString, firstString) == 0 && insideSTATICOBSTACLE)
		{
			sscanf(thisLine, "%*s %s", s1);
			IplImage* img = cvLoadImage(s1);
			if(!img)
				printf("WARNING: Could not load image file: %s\n", s1);
			else
			{
				float ratio = __min( (float)EnvWidth/(float)img->width , (float)EnvHeight/(float)img->height );
				int inv_ratio = round(1.0/ratio);
				CvSize size = cvSize(round(ratio*(float)img->width),round(ratio*(float)img->height));
				IplImage* mono_thresh = cvCreateImage(size, 8, 1);
				mono_thresh->origin = 0;
				int step_mono = mono_thresh->width;
				int channels = img->nChannels;
				int step = img->width * channels;
				int i, j, ii, jj;
				
				for(i=0;i<mono_thresh->height;i++)
					for(j=0;j<mono_thresh->width;j++)
						mono_thresh->imageData[i*step_mono+j*1+0] = 0;
				
				for(a=0; a<img->height; a++)
					for(b=0; b<img->width; b++)
					{
						i = __min(mono_thresh->height-1,__max(0,floor((float)a*ratio)));
						j = __min(mono_thresh->width-1,__max(0,floor((float)b*ratio)));
						if ((int)(*((unsigned char*)img->imageData+a*step+b*channels+0)) + (int)(*((unsigned char*)img->imageData+a*step+b*channels+1)) + (int)(*((unsigned char*)img->imageData+a*step+b*channels+2)) <= 3*127)
							mono_thresh->imageData[i*step_mono+j*1+0] = __min(255,mono_thresh->imageData[i*step_mono+j*1+0]+1);
						if (mono_thresh->imageData[i*step_mono+j*1+0] >= inv_ratio)
						{
							ii = __min(EnvWidth-1,__max(0,j));
							jj = __min(EnvHeight-1,__max(0,EnvHeight-i-1));
							Grid2D[ii][jj] = CELL_OCCUPIED;
							mono_thresh->imageData[i*step_mono+j*1+0] = 255;
						}
						
					}
				//cvNamedWindow("win1",1);
				//cvShowImage("win1", mono_thresh);
				//cvWaitKey();
			}
			continue;
		}
		
		strcpy(checkString, "END_STATICOBSTACLE");
		if(strcmp(checkString, firstString) == 0 && insideSTATICOBSTACLE)
		{
			insideSTATICOBSTACLE = false;
			continue;
		}
		
		// +++++++++++++++++++++++++++++++++++++++++++
		
		// -END_FILE-
		strcpy(checkString, "-END_FILE-");
		if(strcmp(checkString, firstString) == 0)
		{
			FileEnd = true;
			// Set computed values
			
			continue;
		}
		
	}
}

int VoronoiEnv_t::Length2Discritization_i(float x) { return round( x / LENGTH_SCALE ); }
float VoronoiEnv_t::Length2Discritization_f(float x) { return ( x / LENGTH_SCALE ); }


