/*
******************************************************************************************
*                                                                                        *
*    Code for                                                                            *
*    Distributed Coverage and Exploration in Unknown Non-Convex Environments             *
*    Copyright (C) 2010  Subhrajit Bhattacharya                                          *
*                                                                                        *
*    This program is free software: you can redistribute it and/or modify                *
*    it under the terms of the GNU General Public License as published by                *
*    the Free Software Foundation, either version 3 of the License, or                   *
*    (at your option) any later version.                                                 *
*                                                                                        *
*    This program is distributed in the hope that it will be useful,                     *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of                      *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                       *
*    GNU General Public License for more details <http://www.gnu.org/licenses/>.         *
*                                                                                        *
*                                                                                        *
*    Contact: subhrajit@gmail.com, http://fling.seas.upenn.edu/~subhrabh/                *
*                                                                                        *
*                                                                                        *
******************************************************************************************
*/

#define FLOAT_INF 1.0/0.0

#define CELL_OCCUPIED (INT_MAX)
#define CELL_NOT_OCCUPIED 0
#define CELL_OCCUPIED_THRESH (INT_MAX-1)
#define LOG_HALF 0.693147181

class point2D_t
{
public:
	int x;
	int y;
	void set(int xx, int yy) { x=xx; y=yy; };
	float dist(point2D_t p) {return sqrt((float)((x-p.x)*(x-p.x)+(y-p.y)*(y-p.y))); };
};

point2D_t CreatePoint2D (int xx, int yy);

class color3_t
{
public:
	float r, g, b;
	void set(float rr, float gg, float bb) { r=rr; g=gg; b=bb; };
	void set(float ii) { r=ii; g=ii; b=ii; };
};

// =====================================================

class Robot_t
{
public:
	float** prob_sum_map; // May point to prob_map_g if a centralized probability map is maintained
	int** measurement_count_map; // May point to prob_map_g if a centralized probability map is maintained
	float** entropy_map;
	float** entropy_smooth_map;
	int** occupancy_map; // May point to occupancy_map_g if a centralized occupancy map is maintained
	int** valid_target_map;
	bool recomputeTarget;
	
	point2D_t coord;
	SBPL2DGridSearch* search_expansion;
	point2D_t TargetCoord;
	vector<point2D_t> InverseTrajectory;
	vector<point2D_t> TrajHistory;
	
	float sensor_rad_sq;
	float sensor_near_truth, sensor_far_truth;
};

class VoronoiEnv_t
{
public:

	int EnvWidth, EnvHeight;
	int** Grid2D;
	unsigned char** VoronoiTes; // VoronoiTes[x][y]=RobotIndex+1 (for x,y belonging to RobotIndex); VoronoiTes[x][y] = 0 (for obstacles)
	
	float** prob_sum_map_g;
	int** measurement_count_map_g;
	float** entropy_map_g;
	float** entropy_smooth_map_g;
	int** occupancy_map_g;
	
	int ROBOT_COUNT;
	vector<Robot_t> robots;
	
	bool SHARE_MAP, MAP_NO_UNCERTAINTY;
	bool DONT_DRAW_TRAJECTORY, DONT_DRAW_HISTORY;
	int DRAW_INTERVAL;
	vector<int> CAPTURE_PNG_FRAMES;
	float P_THRESH_OBS;
	int ENTROPY_SMOOTHEN_RADIUS;
	float ENTROPY_WEIGHT_POWER;
	float ENTROPY_THRESH_FOR_UNCERTAINTY;
	float TARGET_CELL_MIN_ENTROPY;
	float METRIC_WEIGHT_OF_ENTROPY;
	float LENGTH_SCALE;
	bool TARGET_NEIGHBOUR_GRAD;
	
	VoronoiEnv_t() { ROBOT_COUNT=0; };
	void ReadEnvFile(const char* sEnvFile); // Calls MemAlloc(), AddRobot()
	int Length2Discritization_i(float x);
	float Length2Discritization_f(float x);
	void MemAlloc(void);
	void AddRobot(int x, int y, float sensor_Rad_sq, float sensor_Near_truth, float sensor_Far_truth);
	
	void UpdateMaps(void);
	void DoExpansions(void);
	void CreateTessellations(void);
	void ComputeTargets(void);
		point2D_t ProjectPointInsideTessellation(point2D_t pt, int rob_index);
		point2D_t ProjectPointInsideHighEntropy(point2D_t pt, int rob_index, float entropy_thresh);
	void ComputeTargets_TARGET_NEIGHBOUR_GRAD(void); // Experimental
	void MoveRobots(void);
	
	bool IsCellVisibleToRobot(int cell_x, int cell_y, int rob_index);
	
	float PlotScale;
};

// ------------------------------------------

float PowerDiagramDistance(float dist, float RadSq);
float SensorProbabilityOfTrue(float DistSq, float SensorRadSq, float SensorNearTruth, float SensorFarTruth);
bool SensorSample(bool TrueOccupancy, float ProbabilityOfTrue);

